#include "troxengine.h"
#include "troxfont.h"

#define DEBUG(x) 


typedef struct {
    signed short int dp,z;
    char c;
} Smallballpixel;

typedef struct {
    signed short int x,y,z;
    char c;
} Textballpixel;

typedef struct {
    int r,g,b;
} Palette;

int smallballradius;

short signed int * bigballz;
unsigned char * bigballimage;
short signed int * zbuffer;
unsigned char * image;
unsigned char * statusarea;

Smallballpixel * smallball;
int smallballsize;

Textballpixel * textball1, * textball2;
unsigned int textball1size, textball2size;

Palette savepal[16];
Palette pal[256];


void displayimage(void)
{
    register int line;
    unsigned char * im;
    im=image;
    for (line=0;line<HIGHT;line++) {
	vga_drawscanline(line,im);
	im+=WIDTH;
    }
}


void displaystatus(void) 
{
    register int line;
    for (line=0;line<STATUSHIGHT;line++) {
	vga_drawscanline(line+HIGHT,&statusarea[line*WIDTH]);
    }
}


void clearimage(void)
{
    memset((void*)zbuffer,128,WIDTH*HIGHT*sizeof(short int));
    memcpy((void*)image,(void*)bigballimage,WIDTH*HIGHT);
}


void clearstatusarea(void)
{
    memset((void*)statusarea,0,WIDTH*STATUSHIGHT*sizeof(char));
}


void setpalette(const int & i, const int & r, const int & g, const int & b)
{
    pal[i].r=r PALECOLOR;
    pal[i].g=g PALECOLOR;
    pal[i].b=b PALECOLOR;
}


void generatetextballs(int radius1,int radius2)
{
    signed int x,y;
    float dist;
    textball1=(Textballpixel *) malloc((radius1*2+1)*(radius1*2+1)*sizeof(Textballpixel));
    if (textball1==NULL) {
	printf("troxengine: Could not get enough memory!\n");
	exit(1);
    }
    textball1size=0;
    for (y=-radius1;y<=radius1;y++) {
	for (x=-radius1;x<=radius1;x++) {
	    dist=sqr(float(x))+sqr(float(y));
	    if (dist<=sqr(radius1)) {
		textball1[textball1size].x=x;
		textball1[textball1size].y=y;
		textball1[textball1size].z=(signed short int)(sqrt(sqr(radius1)-dist));
		textball1[textball1size].c=
		    (unsigned char)((1-sqrt(sqr(radius1)-dist)/radius1)*15+float(random()/float(RAND_MAX)));
		textball1size++;
	    }
	}
    }
    textball2=(Textballpixel *) malloc((radius2*2+1)*(radius2*2+1)*sizeof(Textballpixel));
    if (textball2==NULL) {
	printf("troxengine: Could not get enough memory!\n");
	exit(1);
    }
    textball2size=0;
    for (y=-radius2+1;y<=radius2;y++) {
	for (x=-radius2+1;x<=radius2;x++) {
	    dist=sqr(float(x)+0.5)+sqr(float(y)+0.5);
	    if (dist<=sqr(radius2)) {
		textball2[textball2size].x=x;
		textball2[textball2size].y=y;
		textball2[textball2size].z=0;
		textball2[textball2size].c=
		    (unsigned char)((1-sqrt(sqr(radius2)-dist)/radius2)*60);
		textball2size++;
	    }
	}
    }
}


void setsmallballradius(int radius)
{
    signed int x,y,p,np;
    float dist;
    if (radius>MAXRADIUS) {radius=MAXRADIUS;}
    if (radius<MINRADIUS) {radius=MINRADIUS;}
    smallballradius=radius;
    smallball=(Smallballpixel *) 
	malloc((smallballradius*2+1)*(smallballradius*2+1)*sizeof(Smallballpixel));
    if (smallball==NULL) {
	printf("troxengine: Could not get enough memory!\n");
	exit(1);
    }
    smallballsize=0;
    p=0;
    for (y=-smallballradius;y<=smallballradius;y++) {
	for (x=-smallballradius;x<=smallballradius;x++) {
	    dist=sqr(float(x))+sqr(float(y));
	    if (dist<=sqr(smallballradius)) {
		np=(y*WIDTH+x);
		smallball[smallballsize].dp=p-np;
		p=np;
		smallball[smallballsize].z=(signed short int)(sqrt(sqr(smallballradius)-dist));
		smallball[smallballsize].c=
		    (unsigned char)((1-sqrt(sqr(smallballradius)-dist)/
				     smallballradius)*15+float(random()/float(RAND_MAX)));
		smallballsize++;
	    }
	}
    }
}


void outtextxy(signed int x, signed int y, char * text)
{
    unsigned int i,k,l,m;
    signed int xx,yy;
    for (i=0; i<strlen(text);i++) {
	if ((text[i]>32)&&(text[i]<96)) {
	    for (k=0; k<5; k++) {
		for (l=0;l<7;l++) {
		    if ((font[(text[i]-33)*6+k]>>l)&1==1) {
			for (m=0;m<textball2size;m++) {
			    xx=x+(i*6+k)*(2*TEXTBALL2RADIUS)+textball2[m].x+TEXTBALL2RADIUS;
			    yy=y+l*(2*TEXTBALL2RADIUS)+textball2[m].y+TEXTBALL2RADIUS;
			    if ((xx>=0)&&(xx<WIDTH)&&(yy>=0)&&(y<STATUSHIGHT)) {
				statusarea[yy*WIDTH+xx]=textball2[m].c;
			    }
			}
		    }
		}
	    }
	}
    }
}


void createpalette(void)
{
    int i;
    setpalette(0,42,42,42);
    for (i=1;i<42;i++) {
	setpalette(i,42-i,42-i,42);
    }
    for (i=0; i<16;i++) {
	// red balls:
	setpalette(128      +i,56,32-i*2,32-i*2);
	// hidden:
	setpalette(128+16   +i,36,20-i  ,32-i  );
	// green balls:
	setpalette(128   +32+i,32-i*2,44,32-i*2);
	// hidden:
	setpalette(128+16+32+i,20-  i,30,32-i  );
	// yellow balls:
	setpalette(128   +64+i,44,  44  ,32-i*2);
	// hidden:
	setpalette(128+16+64+i,30,  30  ,32-i  );
	// orange balls:
	setpalette(128   +96+i,56,  42-i,32-i*2);
	// hidden:
	setpalette(128+16+96+i,36,30-i/2,30-i  );
    }
}


void setblack(const int & c)
{
    register int i;
    for (i=0; i<16; i++) {
   	// black ball:
	setpalette(128      +i+(c<<5),48-i*2,48-i*2,48-i*2);
	// hidden:
	setpalette(128+16   +i+(c<<5),20-i,20-i,32-i);
    }
}


void doneengine(void)
{
    int i;
    vga_setmode(TEXT);
    printf("Shutting down TROX-engine...\n");
    for (i=0;i<16;i++) {vga_setpalette(i,savepal[i].r,savepal[i].g,savepal[i].b);}
}


void initengine()
{
    signed int i,k,x,y;
    double dist;
    unsigned char c;
    signed short int h;
    vga_init();
    printf("\n");
    printf("\n");
    printf("Initializing TROX-engine!\n");
    if (atexit(doneengine)<0) {perror("troxengine:");}
    generatetextballs(TEXTBALL1RADIUS,TEXTBALL2RADIUS);
    bigballz=(short signed int *)malloc(WIDTH*HIGHT*sizeof(short int));
    bigballimage=(unsigned char *)malloc(WIDTH*HIGHT*sizeof(char));
    zbuffer=(short signed int*)malloc(WIDTH*HIGHT*sizeof(short int));
    image=(unsigned char*)malloc(WIDTH*HIGHT*sizeof(char));
    statusarea=(unsigned char *)malloc(WIDTH*STATUSHIGHT*sizeof(char));
    if ((bigballz==NULL)||(bigballimage==NULL)||
	(zbuffer==NULL)||(image==NULL)||(statusarea==NULL)) {
	printf("troxengine: Could not get enough memory!\n");
	exit(1);
    }
    clearstatusarea();
    memset((void*)bigballz,128,WIDTH*HIGHT*sizeof(short int));
    memset((void*)bigballimage,0,WIDTH*HIGHT);
    clearimage();
    for (y=0;y<HIGHT;y++) {
	for (x=0;x<WIDTH/2;x++) {
	    dist=sqr(float(x-WIDTH/4))+sqr(float(y-HIGHT/2));
	    if (dist<=sqr(BIGBALLRADIUS)) {
		c=(unsigned char)((1-sqrt(sqr(BIGBALLRADIUS)-dist)/
				   BIGBALLRADIUS)*40+1+float(random()/float(RAND_MAX)));
		bigballimage[y*WIDTH+x]=c;
		bigballimage[y*WIDTH+WIDTH-x]=c;
		h=(signed short int)(sqrt(sqr(BIGBALLRADIUS)-dist));
		bigballz[y*WIDTH+x]=h;
		bigballz[y*WIDTH+WIDTH-x]=h;
	    }
	}
    }
    setsmallballradius(SMALLBALLRADIUS);
    for (i=0;i<16;i++) {
	vga_getpalette(i,&savepal[i].r,&savepal[i].g,&savepal[i].b);
    }
    for (k=42;k>=0;k--) {
	for (i=1;i<16;i++) {vga_setpalette(i,k,k,k);}
	usleep(20000);
    }
    if (vga_setmode(G640x480x256)) {
        vga_setmode(TEXT);
	printf("troxengine: Could not initialize graphics!\n"); 
	exit(1);
    }
    for (i=0;i<255;i++) {
	vga_setpalette(i,0,0,0);
    }
    createpalette();
}


void drawsmallball(signed short const & x, signed short const & y, signed short const & z, 
		   unsigned char const & c, int const & side)
{
    register unsigned short i;
    register signed int zz;
    register signed int p;
    Smallballpixel * ball;
    ball=smallball;
    p=(y+(HIGHT>>1))*WIDTH+(x+(WIDTH>>2)+side*(WIDTH>>1));
    for (i=0;i<smallballsize;i++) {
	zz=z+(*ball).z;
	p+=(*ball).dp;
	if (zbuffer[p]<zz) {
	    zbuffer[p]=zz;
	    if (bigballz[p]<zz) {
		image[p]=(*ball).c|128|(c<<5);
	    }
	    else {
		image[p]=(*ball).c|144|(c<<5);
	    }
	}
	ball++;
    }
}


void drawdemoball(signed short const & x, signed short const & y, signed short const & z, 
		   unsigned char const & c)
{
    register unsigned short i;
    register signed int zz;
    register signed int p;
    Smallballpixel * ball;
    ball=smallball;
    p=y*WIDTH+x;
    for (i=0;i<smallballsize;i++) {
	zz=z+(*ball).z;
	p+=(long signed)(*ball).dp;
	if (zbuffer[p]<zz) {
	    zbuffer[p]=zz;
	    if (bigballz[p]<zz) {
		image[p]=(*ball).c|128|(c<<5);
	    }
	    else {
		image[p]=(*ball).c|144|(c<<5);
	    }
	}
	ball++;
    }
}


void showpal(void)
{
    vga_setpalvec(0,255,(int*)pal);
}


void fadepal(const int & fade)
{
    int i;
    for (i=0;i<256;i++) {vga_setpalette(i,pal[i].r*fade/50,pal[i].g*fade/50,pal[i].b*fade/50);}
}


void fadetogrey(int fade)
{
    int i;
    for (i=0;i<256;i++) {
	vga_setpalette(i,(pal[i].r*fade+42*(50-fade))/50,
		       (pal[i].g*fade+42*(50-fade))/50,
		       (pal[i].b*fade+42*(50-fade))/50);
    }
}


void fadein(void)
{
    int i,k;
    for (i=0; i<=42; i++) {
	for (k=0; k<256; k++) {vga_setpalette(k,i,i,i);}
    }
}


void greyin(void)
{
    int i;
    for (i=0; i<=50; i++) {fadetogrey(i);}
}


void greyout(void)
{
    int i;
    for (i=50; i>=0; i--) {fadetogrey(i);}
}


int getsmallballradius(void)
{
    return smallballradius;
}


void fadeout(void) 
{
    int i,k;
    for (i=42; i>=0; i--) {
	for (k=0; k<256; k++) {vga_setpalette(k,i,i,i);}
    }
}

